(*
 *  beware IRC services, bparse.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  some modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

 {most P10 stuff is to be in this unit for easy porting to another server protocol}

unit bparse;

interface

uses bservicemngr , buser , pgdebugout , bchannel ,breplies;

var
  isconnected:boolean;
  online:boolean;
  mynum:string;
  hostservername:string;
  sendqsize:integer;

procedure send(const s:string);
procedure sendflush;
procedure sendburst;
procedure onconnected;
procedure parsecommand(mine:boolean;const s:string);
procedure sendbot_nick(us:tuser);
procedure sendparse(const s:string);
procedure sendkill(target:tuser;const fromstr,reason:string);

{
i wrap everything so a bot can without much effort be used on a different server-server protocol
}
function sendprefixstr(from:tuser):string;
function sendtalkstr(from,target:tuser):string;
procedure bot_join(sptr:tuser;channame:string;modes,modeparams:string;ts:integer);
procedure bot_quit(sptr:tuser;reason:string);

procedure onnewuser(us:tuser);
procedure onleavechannel(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
procedure onnewchannel(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
procedure onchanneltschange(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
procedure onjoinchannel(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
procedure ontopic(channel:tchannel;source:tuser;var bounce:boolean);
function botmodestring(bot:tuser;ch:tchannel) :string;
procedure onsendeob;
function sprefix(us:tuser;const cmd:string):string;
function cmdstr(num:integer):string;
procedure sendnosuchserver(sptr:tuser;server:string);
procedure sendreply(sptr:tuser;reply:integer;bulk:string);

implementation

uses
  bservunit  , bstuff     , sysutils , btime  , bconsts , bcmds       ,
  bircrandom , bconfig    , bserver  , bcrash , b_burst , b_clearmode ,
  bipcheck   , binipstuff , b_whois  ;

function sendtalkstr(from,target:tuser):string;
begin
  result := sendprefixstr(from)+TOK_NOTICE+' '+target.idstr+' :';
end;

{gives the ":sender " part of a message to send}
function sendprefixstr(from:tuser):string;
begin
  result := from.idstr+' ';
end;

procedure sendbot_nick(us:tuser);
var
  iptemp:string;
begin
  {$ifdef ipv6}
  if us.binip.family = AF_INET6 then begin
    iptemp :=
    p10inttostr(htonl(us.binip.ip6.u6_addr32[0]),6)+
    p10inttostr(htonl(us.binip.ip6.u6_addr32[1]),6)+
    p10inttostr(htonl(us.binip.ip6.u6_addr32[2]),6)+
    p10inttostr(htonl(us.binip.ip6.u6_addr32[3]),6)
  end else
  {$endif}
  begin
    iptemp := p10inttostr(htonl(us.binip.ip),6);
  end;
  send(mynum+' '+TOK_NICK+' '+us.name+' 1 '+inttostr(us.ts)+' '+us.userid+' '+us.host+usermodestr(us)+' '+iptemp+' '+us.idstr+' :'+us.fullname);
end;

procedure sendkill(target:tuser;const fromstr,reason:string);
begin
  sendparse(mynum+' '+TOK_KILL+' '+target.idstr+' :'+me.name+'!'+fromstr+' ('+reason+')');
end;

procedure sendflush;
begin
  if not isconnected then exit;
  sock.send(nil,0);
  sendqsize := 0;
end;

procedure send(const s:string);
begin
  if isconnected then begin
    inc(sendqsize,length(s)+2);
    sock.putstringinsendbuffer(s+#13#10);
{    if sendqsize > 1500 then sendflush;}
  end;
end;

function channelhaskuser(ch : tchannel) : boolean;
var
  uc : tuserchan;
begin
  result := false;
  uc := tuserchan(ch.user);
  while assigned(uc) do begin
    if (uc.us.modeflag and usermode_service)=usermode_service then begin
      result := true;
      exit;
    end;
    uc := tuserchan(uc.next2);
  end;
end;

procedure bot_join(sptr:tuser;channame:string;modes,modeparams:string;ts:integer);
var
  ch  : tchannel;
  ts2 : longint;
begin
  // the reason this is so complex is because we have to aviod USING burst
  // messsages outside of burst so we can run without a U: line

  ch := findchan(channame);
  if assigned(ch) and (ch.ts>0) and (ts<ch.ts) then if not channelhaskuser(ch) then begin
    // we don't do any of our netrider stuff to channel that have a +k on them
    if (pos('i',modes) <> 0) or (pos('k',modes) <> 0) then begin
      // we have to be carefull because kicking the last user will invalidate ch
      //first reduce the channel to 1 user
      while ch.usercount >1 do begin
        sendparse(me.idstr+' '+tok_kick+' '+ch.name+' '+tuserchan(ch.user).us.idstr+' :net rider');
      end;
      // then kick that user destroying the channel
      sendparse(me.idstr+' '+tok_kick+' '+ch.name+' '+tuserchan(ch.user).us.idstr+' :net rider');
    end else begin
      sendparse(me.idstr+' '+TOK_CLEARMODE+' '+ch.name+' '+schanmodesupported);
      sendparse(me.idstr+' '+TOK_TOPIC+' '+ch.name+' :');
    end;
  end;

  ch := findchan(channame);
  if ts =0 then ts := irctime;
  ts2 := ts;
  if ch <> nil then ts := ch.ts;
  sendparse(sptr.idstr+' '+TOK_JOIN+' '+channame+' '+inttostr(ts));
  if ts2 > ts then ts2 := ts;
  if length(modes)>0  then sendparse(me.idstr+' '+TOK_MODE+' '+channame+' '+modes+' '+modeparams+' '+inttostr(ts2));
end;

procedure bot_quit(sptr:tuser;reason:string);
begin
  sendparse(sptr.idstr+' '+TOK_QUIT+' :'+reason);
end;


procedure onconnected;
begin
  send('PASS '+opt.connectpass);
  send('SERVER '+me.name+' 1 '+inttostr(irctime)+' '+inttostr(irctime)+' J10 '+mynum+p10inttostr(maxbot,CCClen)+' :'+opt.info);
end;

procedure sendburst;
var
  a,c:integer;
  ch:tchannel;
  uc:tuserchan;
  ban:tban;
  bool,goed:boolean;
  s:string;
begin
  {send all bots already in list}
  for a := 0 to maxbot do if botarray[a] <> nil then begin
    sendbot_nick(botarray[a]);
  end;

  {sendburst all currently existing channels}
  ch := tchannel(globalchanlist);
  while ch <> nil do begin

      s := '';
      uc := tuserchan(ch.user);
      ban := ch.banlist;
      goed := true;

      repeat
        if goed then begin
          s := chanmodestr(ch,true);
          goed := false;
        end else s := '';
        c := 0;
        bool := true;
        while (uc <> nil) and (length(s) < 400) do begin
          if bool then begin
            s := s + ' ';
            bool := false;
          end else s := s + ',';
          s := s + uc.us.idstr;
          if uc.flags and (userchanmodeflagmask) <> c then begin
            c := uc.flags and (userchanmodeflagmask);
            s := s + ':';
            for a := maxuserchanmodetable downto 0 do begin
              if flag_isset(c,userchanmodetable[a].flag) then s := s + userchanmodetable[a].c;
            end;
          end;
          uc := tuserchan(uc.next2);
        end;
        bool := true;
        while (ban <> nil) and (length(s) < 400) do begin
          if bool then begin
            s := s + ' :%';
            bool := false;
          end;
          s := s + ban.mask+' ';
          ban := tban(ban.next);
        end;
        send(me.idstr+' '+TOK_BURST+' '+ch.name+' '+inttostr(ch.ts)+s);
      until (uc = nil) and (ban = nil);

    ch := tchannel(ch.next);
  end;

  {call onsendburst of all modules}

  send(mynum+' '+TOK_END_OF_BURST);
  onsendeob;
end;


{
send something, and parse it - the bots see the message being sent
(a NOTICE can be sent w/o parsing it)

cpu consuming but easy way to keep everything synched

example:

chanserv -> MODE change
scanner -> PRIVMSG operserv GLINE
}
procedure sendparse(const s:string);
begin
  send(s);
  parsecommand(true,s);
end;

procedure onnewuser(us:tuser);
var
  a:integer;
  s:tserviceunit;
begin
  ipcheck_remoteclient(us);
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.onnewuser <> nil then s.onnewuser(us,s.tag);
  end;
end;

procedure onsendeob;
var
  a:integer;
  s:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.onsendeob <> nil then s.onsendeob(s.tag);
  end;
end;

procedure onnickchange(us:tuser);
var
  a:integer;
  s:tserviceunit;
begin
  //ipcheck_remoteclient(us);
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.onnickchange <> nil then s.onnickchange(us,s.tag);
  end;
end;


procedure onaccount(sptr:tuser;us:tuser);
var
  a:integer;
  s:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.onaccount <> nil then s.onaccount(sptr,us,s.tag);
  end;
end;

procedure onsignonts(sptr:tuser;us:tuser);
var
  a:integer;
  s:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.onsignonts <> nil then s.onsignonts(sptr,us,s.tag);
  end;
end;


procedure onchanmsg(sptr:tuser;target:tchannel;s:string);
var
  a:integer;
  currentserviceunit:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    currentserviceunit := serviceunitarray[a];
    // note i do cheap pointer nil tests before expensive srvunitisonchannel test
    if (currentserviceunit <> nil)  then if @currentserviceunit.onchanmsg <> nil then if srvunitisonchannel(currentserviceunit,target) then currentserviceunit.onchanmsg(sptr,target,s,currentserviceunit.tag);
  end;
end;

procedure onleavechannel(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
var
  a:integer;
  currentserviceunit:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    currentserviceunit := serviceunitarray[a];
    // note i do cheap pointer nil tests before expensive srvunitisonchannel test
    if (currentserviceunit <> nil)  then if @currentserviceunit.onleavechannel <> nil then {if srvunitisonchannel(currentserviceunit,channel) then} currentserviceunit.onleavechannel(channel,source,target,flags,message,currentserviceunit.tag);
  end;
end;

procedure onjoinchannel(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
var
  a:integer;
  currentserviceunit:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    currentserviceunit := serviceunitarray[a];
    // note i do cheap pointer nil tests before expensive srvunitisonchannel test
    if (currentserviceunit <> nil)  then if @currentserviceunit.onjoinchannel <> nil then {if srvunitisonchannel(currentserviceunit,channel) then} currentserviceunit.onjoinchannel(channel,source,target,flags,message,currentserviceunit.tag);
  end;
end;

procedure onnewchannel(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
var
  a:integer;
  currentserviceunit:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    currentserviceunit := serviceunitarray[a];
    // note i do cheap pointer nil tests before expensive srvunitisonchannel test
    if (currentserviceunit <> nil)  then if @currentserviceunit.onnewchannel <> nil then {if srvunitisonchannel(currentserviceunit,channel) then} currentserviceunit.onnewchannel(channel,source,target,flags,message,currentserviceunit.tag);
  end;
end;

procedure onchanneltschange(channel:tchannel;source:tuser;target:tuser;flags:byte;message:string);
var
  a:integer;
  currentserviceunit:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    currentserviceunit := serviceunitarray[a];
    // note i do cheap pointer nil tests before expensive srvunitisonchannel test
    if (currentserviceunit <> nil) then if @currentserviceunit.onchanneltschange <> nil then {if srvunitisonchannel(currentserviceunit,channel) then} currentserviceunit.onchanneltschange(channel,source,target,flags,message,currentserviceunit.tag);
  end;
end;

procedure ontopic(channel:tchannel;source:tuser;var bounce:boolean);
var
  a:integer;
  currentserviceunit:tserviceunit;
begin
  for a := maxserviceunit downto 0 do begin
    currentserviceunit := serviceunitarray[a];
    // note i do cheap pointer nil tests before expensive srvunitisonchannel test
    if (currentserviceunit <> nil)  then if @currentserviceunit.ontopic <> nil then {if srvunitisonchannel(currentserviceunit,channel) then} currentserviceunit.ontopic(channel,source,bounce,currentserviceunit.tag);
    if bounce then exit;
  end;
end;


procedure parsecommand(mine:boolean;const s:string);
var
  a,b,c,d:integer;
  cmd,cmd2:string;
  parc:integer;
  newcommand:boolean;
  parv,parv2:tparams;
  bool:boolean;
  sender:string;
  s1:string;
  sptr:tuser;
  us:tuser;
  ch:tchannel;
  uc:tuserchan;
  p:pointer;
  newchannel : boolean;
  bouncekick : boolean;
  ipstr:string;
begin
 try

  debugout('parsecommand entered');

  if s = '' then exit;
  if length(s) > maxmessagelength then begin
    send(me.idstr+' '+tok_wallops+' :'+'received too long line: '+inttostr(length(s))+' ['+copy(s,1,50)+']');
    exit;
  end;
  {fs.memo1.lines.add(s);}

  if s[1] = ':' then begin
    newcommand := false;
    a := pos(' ',s);
    sender := copy(s,2,a-2);
  end else begin
    if (online or mine) then begin
      newcommand := true;
      a := pos(' ',s);
      sender := copy(s,1,a-1);
    end else begin
      newcommand := false;
      sender := '';
      a := 1;
    end;
  end;
  if a = 0 then exit;
  if sender <> '' then sender := convertidstr(sender);
  if sender <> '' then if (a >= length(s)) then exit;

  {a is index in parsed string - get index of start of command}
  while s[a] = ' ' do inc(a);
  b := 0;
  c := length(s);
  bool := false;
  while (a <= c) do begin
    if b > mparams then break;
    if s[a] = ':' then begin
      bool := true;
      break;
    end;
    d := a;
    while (s[a] <> ' ') and (a <= c) do inc(a);
    parv[b] := copy(s,d,a-d);
    inc(b);
    while (s[a] = ' ') and (a <= c) do inc(a)
  end;
  inc(a);
  if bool then begin
    parv[b] := copy(s,a,c);
    inc(b)
  end;

  parc := b;
  if parc = 0 then exit; {no command, only a :sender prefix}

  cmd2 := parv[0];
  cmd := ircupper(cmd2);

  if not mine then if cmd = MSG_ERROR then begin
    {fs.caption := parv[parc-1];}
    if pos('throttled',parv[parc-1]) <> 0 then begin
      reconnectcountdown := 45;
    end;
  end;

  {unknown command}

  if not mine then if copy(sender,1,SSlen) = me.idstr then exit;
  if not (mine or online) then begin
    if cmd = 'SERVER' then begin
      us := adduser;
      hostservuser := us;
      setnickname(us,parv[1]);
      us.hops := 1;
      us.fullname := copy(parv[8],1,maxgcoslength);
      parv[6] := convertidstr(parv[6]);

      assert((length(parv[6]) = SSCCClen) and (p10strtoint(parv[6]) >= 0) ,'parse: SERVER: invalid numeric',s);

      us.idstr := copy(parv[6],1,SSlen);

      assert(p10server[p10strtoint(us.idstr)] = nil,'parse: SERVER: server numeric collision',s);
      addserver(us,p10strtoint(us.idstr),p10strtoint(copy(parv[6],1+SSlen,CCClen)),me.server);
      setflag(us.server.flags,servflag_joining);
      online := true;
      sendburst;
    end;
    exit;
  end;

  if newcommand then begin
    {prefix had no colon, its a numeric}
    sptr := findnumeric(sender);
  end else begin
    {prefix had a colon, it's a name}
    sptr := findname(sender);
  end;

  if sptr = nil then exit;
  if cmd = TOK_PING then begin
    send(mynum+' '+TOK_PONG+' '+me.idstr+' :'+parv[1]);
  end else if cmd = TOK_ACCOUNT then begin
    us := findnumeric(parv[1]);

    {assert(assigned(us),'parse: ACCOUNT: user not found',s);}
    if assigned(us) then begin
      if parv[2] <> '' then begin
        us.account := parv[2];
        onaccount(sptr,us);
      end;
    end;
  end else if cmd = TOK_NICK then begin
    if isserver(sptr) then begin
      {server introducing new client}

      parv[parc-2] := convertidstr(parv[parc-2]);
      assert((length(parv[parc-2]) = SSCCClen) and (p10strtoint(parv[parc-2]) >= 0) ,'parse: NICK: invalid numeric',s);

      a := p10strtoint(copy(parv[parc-2],1,SSlen));
      b := p10strtoint(copy(parv[parc-2],1+SSlen,CCClen));
      assert((a >= 0) and (b >= 0),'parse: N: bogus numeric',s);
      assert(p10server[a] <> nil,'parse: N: servernum does not exist',s);
      b := b and p10server[a].p10max;

      assert(p10server[a].p10slots[b] = nil,'parse: N: numeric collision',s);

      {nick collision?}
      us := findname(parv[1]);
      if us <> nil then begin
        assert(us.server = me.server,'parse: N: nick collision, both users are not a service',s);
        exit;
      end;

      us := adduser;
      p10server[a].p10slots[b] := us;
      us.p10slotnum := b;
      setnickname(us,parv[1]);
      us.ts := strtointdef(parv[3],0);
      us.hops := strtointdef(parv[2],0);
{      assert(us.hops = sptr.hops,'parse: N: client hopcount isn''t same as server hopcount',s);}
      us.userid := parv[4];
      us.host := parv[5];
      us.uhost := ircupper(parv[6]);

      us.idstr := parv[parc-2];

      ipstr := parv[parc-3];
      {$ifdef ipv6}
      if length(ipstr) = 24 then begin
        us.binip.family := AF_INET6;
        us.binip.ip6.u6_addr32[0] := htonl(p10strtoint(copy(ipstr,1,6)));
        us.binip.ip6.u6_addr32[1] := htonl(p10strtoint(copy(ipstr,7,6)));
        us.binip.ip6.u6_addr32[2] := htonl(p10strtoint(copy(ipstr,13,6)));
        us.binip.ip6.u6_addr32[3] := htonl(p10strtoint(copy(ipstr,19,6)));
      end else
      {$endif}
      begin
        us.binip.family := AF_INET;
        a := p10strtoint(ipstr);
        us.binip.ip := htonl(a);
      end;

      us.server := sptr.server;
      setflag(us.flags,userflag_isclient);
      inc(count.globalclients);

      if parv[6,1] = '+' then begin
      {modes}
        for a := 0 to maxusermodetable do if pos(usermodetable[a].c,parv[6]) <> 0 then setflag(us.modeflag,usermodetable[a].flag);
      end;

      if isoper(us) then inc(count.oper);
      if flag_isset(us.modeflag,usermode_invisible) then inc(count.invisible);
      us.fullname := parv[parc-1];

      {also called if it is mine?}
      onnewuser(us);
      if parv[6,1] = '+' then begin
        {search for mode +r (account burst) stores the login name
        and then passes the message on to any interested bots}
        if pos('r',parv[6]) <>0 then begin
          us.account := parv[7];
          onaccount(sptr,us);
        end;
      end;
    end else begin
      {client changing nick}
      setnickname(sptr,parv[1]);
      sptr.ts := strtointdef(parv[2],0);
      onnickchange(sptr);
    end;
  end else
  if cmd = TOK_SVSNICK then begin
    (*
    -- must not parse svsnick, new svsnick code
    us := findnumeric(parv[1]);
    if us = nil then exit;
    us2 := findname(parv[2]);

    assert((us2 = nil) or (us2 = us),'parse: SVSNICK: nick collision',s);
    setnickname(us,parv[2]);
    us.ts := strtointdef(parv[3],0);
    *)
  end else
  if cmd = TOK_QUIT then begin
    ircrandom; {call ircrandom to feed randomness}
    uc := tuserchan(sptr.channel);
    while Assigned(uc) do begin
      onleavechannel(uc.ch,sptr,sptr,lc_quit,parv[parc-1]);
      uc := tuserchan(uc.next);
    end;
    sptr.destroy;
  end else
  if cmd = TOK_KILL then begin
    us := findnumeric(parv[1]);

    if assigned(us) then begin
      if (us.server = me.server) and (us.srvunit <> nil) then begin
        assert(not isserver(sptr),'service '+us.name+' killed by server '+sptr.name,'reason: '+parv[parc-1]);
        p := us.srvunit;
        tserviceunit(p).onshutdown(tserviceunit(p).tag);
        tserviceunit(p).onstartup(tserviceunit(p).tag);
      end else begin
        us.destroy;
      end;
      uc := tuserchan(us.channel);
      while Assigned(uc) do begin
        onleavechannel(uc.ch,sptr,us,lc_quit+lc_kill,parv[parc-1]);
        uc := tuserchan(uc.next);
      end;
    end;
  end else
  if cmd = TOK_SQUIT then begin
    us := findname(parv[1]);
    if us = me then us := hostservuser;
    assert(us <> nil,'parse: SQUIT: no such server',s);
    netsplitquitreason := tuser(us.server.parentserver.us).name+' '+us.name;
    us.destroy;
    netsplitquitreason := '';
  end else
  if cmd = TOK_SERVER then begin
    {introducing new server}
    parv[6] := convertidstr(parv[6]);
    assert((length(parv[6]) = SSCCClen) and (p10strtoint(parv[6]) >= 0) ,'parse: SERVER: invalid numeric',s);

    us := adduser;
    setnickname(us,parv[1]);
    us.hops := strtointdef(parv[2],0);
    us.fullname := copy(parv[parc-1],1,maxgcoslength);
    us.idstr := copy(parv[6],1,SSlen);
    assert(p10server[p10strtoint(us.idstr)] = nil,'parse: S: server numeric collision',s);
    addserver(us,p10strtoint(us.idstr),p10strtoint(copy(parv[6],1+SSlen,CCClen)),sptr.server);
    assert(us.hops = sptr.hops+1,'parse: S: bogus hopcount: '+inttostr(sptr.hops)+' '+inttostr(us.hops),s);
    if parv[5] = 'J10' then setflag(us.server.flags,servflag_joining)
    else setflag(us.server.flags,servflag_burstack);
  end else
  if cmd = TOK_CREATE then begin
    assert(isclient(sptr),'parse: C',s);
    b := strtok(parv[1],',',@parv2);
    for a := 0 to b-1 do begin
      newchannel := false;
      ch := findchan(parv2[a]);
      if not isonchannel(sptr,ch) then begin
        if ch = nil then begin
          newchannel := true;
          ch := createchannel;
          setchanname(ch,parv2[a]);
          if ch.name[1] = '+' then setflag(ch.flags,chanflag_modeless);

        end;
        uc := addusertochannel(sptr,ch);
        if newchannel then onnewchannel(ch,sptr,sptr,jc_create,parv[parc-1]);

        c := strtointdef(parv[2],0);
        if ((c < ch.ts) and (c > 0)) or ((ch.ts = 0) and (c > 0)) then  begin
          ch.ts := c;
          onchanneltschange(ch,sptr,sptr,jc_create,parv[parc-1]);

        end;

        onjoinchannel(ch,sptr,sptr,jc_create,parv[parc-1]);
      end else uc := getuserchan(sptr,ch);
      if not flag_isset(ch.flags,chanflag_modeless) then begin
        setflag(uc.flags,userchanflag_op);
      end;

    end;
  end else
  if cmd = TOK_JOIN then begin
    ircrandom; {call ircrandom to feed randomness}
    if parv[1] = '0' then begin
      while assigned(sptr.channel) do begin
        onleavechannel(tuserchan(sptr.channel).ch,sptr,sptr,lc_part+lc_leaveall,'left all channels');
        deluserfromchannel(sptr,tuserchan(sptr.channel).ch,tuserchan(sptr.channel));
      end;
      exit;
    end;
    assert(isclient(sptr),'parse: J',s);
    b := strtok(parv[1],',',@parv2);
    for a := 0 to b-1 do begin
      ch := findchan(parv2[a]);
      if not isonchannel(sptr,ch) then begin
        newchannel := false;
        if ch = nil then begin
          newchannel := true;
          ch := createchannel;
          setchanname(ch,parv2[a]);
          if ch.name[1] = '+' then setflag(ch.flags,chanflag_modeless);
        end;
        {uc := }addusertochannel(sptr,ch);
        if newchannel then onnewchannel(ch,sptr,sptr,jc_join,parv[parc-1]);
        c := strtointdef(parv[2],0);
        if ((c < ch.ts) and (c > 0)) or ((ch.ts = 0) and (c > 0)) then begin
          ch.ts := c;
          onchanneltschange(ch,sptr,sptr,jc_join,parv[parc-1]);
        end;
        onjoinchannel(ch,sptr,sptr,jc_join,parv[parc-1]);
      end;
    end;
  end else
  if cmd = TOK_PART then begin
    b :=  strtok(parv[1],',',@parv2);
    for a := 0 to b-1 do begin
      ch := findchan(parv2[a]);
      if ch <> nil then begin
        uc := getuserchan(sptr,ch);
        if uc <> nil then begin
          onleavechannel(ch,sptr,sptr,lc_part,parv[parc-1]);
          deluserfromchannel(sptr,ch,uc);
        end;
      end;
    end;
  end else
  if cmd = TOK_KICK then begin
    us := findnumeric(parv[2]);
    ch := findchan(parv[1]);
    uc := getuserchan(us,ch);
    // begin kick ack/bounce code
    // kick bounce
    if assigned(us.srvunit) then begin
      if assigned(tserviceunit(us.srvunit).onquerybouncekick) then begin
        bouncekick := false;
        tserviceunit(us.srvunit).onquerybouncekick(ch,sptr,us,parv[parc-1],tserviceunit(us.srvunit).tag,bouncekick);
        if bouncekick then begin
          send(us.idstr+' '+tok_join+' '+ch.name+' '+inttostr(ch.ts));
          //!!! beware finish this when you have put in userchanmodetable
          //send(me.idstr+' '+tok_mode+' '+ch.name+' '+
          exit;
        end;
      end;
    end;
    // kick ack
    if (us.server = me.server) and (sptr.server <> me.server) then begin
      send(us.idstr+' '+tok_part+' '+ch.name);
    end;
    // end kick ack/bounce code

    if uc <> nil then begin
      onleavechannel(ch,sptr,us,lc_kick,parv[parc-1]);
      deluserfromchannel(us,ch,uc);
    end;
  end else
  if (cmd = TOK_MODE) or (cmd = TOK_OPMODE) then begin
    ch := findchan(parv[1]);
    if ch = nil then begin
      us := findnick(parv[1]);
      if us <> nil then begin
        if cmd = TOK_MODE then begin
          setusermode(us,hostservuser,sptr,parc,@parv,mt_mode);
        end else begin
          setusermode(us,hostservuser,sptr,parc,@parv,mt_opmode);

        end;
      end;
      exit;
    end;
    if cmd = TOK_MODE then begin
      setchanmode(ch,hostservuser,sptr,parc,@parv,mt_mode);
    end else begin
      setchanmode(ch,hostservuser,sptr,parc,@parv,mt_opmode);

    end;
  end else if (cmd = TOK_CLEARMODE) then begin
    m_clearmode(sptr,sptr,parc,@parv);
  end else
  if cmd = TOK_PRIVMSG then begin
    ch := findchan(parv[1]);
    if Assigned(ch) then begin
      onchanmsg(sptr,ch,parv[parc-1]);
    end else begin
      a := pos('@',parv[1]);
      if a <> 0 then begin
        s1 := copy(parv[1],a+1,500);
        if not strcompup(s1,me.name) then begin; {not at my server}
          us := nil;
        end else begin
          s1 := copy(parv[1],1,a-1);
          us := findname(s1);
          //if us = nil then exit; {no such nick}
          bool := true;
        end;
      end else begin
        bool := false;
        us := findnumeric(parv[1]);
        //if us = nil then exit;
      end;
      if us <> nil then begin // no such nick/numeric or user not at this server
        if us.srvunit <> nil then begin; {if this user isnt my service bot, ignore message}
          tserviceunit(us.srvunit).onprivmsg(bool,sptr,us,parv[parc-1], tserviceunit(us.srvunit).tag);
        end;
      end;
    end;
  end else
  if cmd = TOK_NOTICE then begin
    //ch := findchan(parv[1]);
    //if Assigned(ch) then begin
    //  onchanmsg(sptr,ch,parv[parc-1]);
    //end else begin
      a := pos('@',parv[1]);
      if a <> 0 then begin
        s1 := copy(parv[1],a+1,500);
        if not strcompup(s1,me.name) then begin; {not at my server}
          us := nil;
        end else begin
          s1 := copy(parv[1],1,a-1);
          us := findname(s1);
          //if us = nil then exit; {no such nick}
          bool := true;
        end;
      end else begin
        bool := false;
        us := findnumeric(parv[1]);
        //if us = nil then exit;
      end;
      if us <> nil then begin // no such nick/numeric or user not at this server
        if us.srvunit <> nil then begin; {if this user isnt my service bot, ignore message}
          tserviceunit(us.srvunit).onnotice(bool,sptr,us,parv[parc-1], tserviceunit(us.srvunit).tag);
        end;
      end;
    //end;
  end else
  if cmd = TOK_TOPIC then begin
    {
    1=channel
    2/-3=ts
    3/-2=ts
    4/-1=topic
    }
    ch := findchan(parv[1]);
    if not assigned(ch) then exit;
    if parc > 4 then b := strtointdef(parv[parc-3],0) else b := 0;
    if parc > 3 then c := strtointdef(parv[parc-2],0) else c := 0;
    if (b > ch.ts) then exit;
    if (c <> 0) and (c < ch.topictime) then exit;
    if c = 0 then c := irctime;
    ch.topictime := c;
    ch.topic := parv[parc-1];
    ch.topicby := sptr.name;
    bool := false;
    ontopic(ch,sptr,bool);
  end else
  if cmd = TOK_BURST then begin
    m_burst(hostservuser,sptr,parc,@parv);
    {parse burst}
  end else
  if cmd = TOK_END_OF_BURST then begin
    if (sptr = hostservuser) then send(mynum+' '+TOK_EOB_ACK);
    clearflag(sptr.server.flags,servflag_joining);
    //writeln(sptr.name);
  end else
  if cmd = TOK_EOB_ACK then begin
    setflag(sptr.server.flags,servflag_burstack);
  end else if cmd = tok_invite then begin
    us := findname(parv[1]);
    ch := findchan(parv[2]);
    if (us <> nil)and(ch<>nil) then begin
      if us.srvunit <> nil then if @tserviceunit(us.srvunit).oninvite <> nil then tserviceunit(us.srvunit).oninvite(sptr,us,ch,tserviceunit(us.srvunit).tag);
    end;
  end else if cmd = '317' then begin
    //reply with signon timestamp from whois
    us := findname(parv[2]);
    if assigned(us) then begin
      us.signonts := strtointdef(parv[4],0);
      onsignonts(sptr,us);
    end;
    //onsignonts(sptr);
  end else if cmd = tok_version then begin
    send(me.idstr+' 351 '+sptr.idstr+' bewareserv'+versionstr+' '+me.name +' :'+versioninfostr);
  end else if cmd = tok_whois then begin
    m_whois(sptr,parc,@parv);
  end;

  //onparse(sptr,parc,@parv);
  debugout('parsecommand complete');

 except
   on e: exception do begin
     send(me.name+' '+TOK_WALLOPS + ' :exception raised dureing parsecommand type '+e.ClassName+' with message '+e.Message+' while parseing message '+s+#13#10);
   end;
 end;
end;

function botmodestring(bot:tuser;ch:tchannel) :string;
var
  uc : tuserchan;
begin
  result := me.idstr;
  uc := getuserchan(bot,ch);

  if assigned(uc) then begin
    if (uc.flags and userchanflag_op)<>0 then begin
      result := bot.idstr;
    end;
  end;
  result := result + ' '+tok_mode+' '+ch.name;
end;

function sprefix(us:tuser;const cmd:string):string;
begin
  result := us.idstr+' '+cmd+' ';
end;
function cmdstr(num:integer):string;
begin
  if num > 1000 then result := rplnumstr[num-1000]
  else assert(false,'bewareserv does not use numbers for non numeric messages','') //result := cmdtable[num].cmd;
end;

procedure sendnosuchserver(sptr:tuser;server:string);
begin
  //if isserver(sptr.from) then
   sendreply(sptr,ERR_NOSUCHSERVER,'* '+getrpl0(ERR_NOSUCHSERVER))
  //else
  // sendreply(sptr,ERR_NOSUCHSERVER,server+' '+getrpl0(ERR_NOSUCHSERVER));
end;

procedure sendreply(sptr:tuser;reply:integer;bulk:string);
begin
  send(me.idstr+' '+inttostr(reply-1000)+' '+sptr.idstr+' '+bulk);
end;


end.
